"""
Code credits: https://github.com/davabase/whisper_real_time
"""
import io
import speech_recognition as sr
import whisper
import torch
import time

from datetime import datetime, timedelta
from queue import Queue
from time import sleep
from sys import platform
from .asr_utils import Config

def transcribe(asr_config_path):
    """Prints the transcribed text to the console."""
    config = Config(asr_config_path)

    phrase_time = None
    last_sample = bytes()
    data_queue = Queue() # Queue containing the audio data from the callback.

    recorder = sr.Recognizer()
    recorder.energy_threshold = config.energy_threshold
    recorder.dynamic_energy_threshold = False
    
    if 'linux' in platform:
        mic_name = config.default_microphone
        if not mic_name or mic_name == 'list':
            print("Available microphone devices are: ")
            for index, name in enumerate(sr.Microphone.list_microphone_names()):
                print(f"Microphone with name \"{name}\" found")   
            return
        else:
            for index, name in enumerate(sr.Microphone.list_microphone_names()):
                if mic_name in name:
                    source = sr.Microphone(sample_rate=44100, device_index=index)
                    break
    else:
        source = sr.Microphone(sample_rate=44100)
        

    if "whisper" in config.model:
        _, model_type = config.model.split("-")
        if model_type != "large" and not config.non_english:
            model_type = model_type + ".en"
        audio_model = whisper.load_model(model_type)

    record_timeout = config.record_timeout
    phrase_timeout = config.phrase_timeout
    
    transcription = ['']
    
    with source:
        recorder.adjust_for_ambient_noise(source)

    def record_callback(_, audio:sr.AudioData) -> None:
        """
        Threaded callback function to recieve audio data when recordings finish.
        audio: An AudioData containing the recorded bytes.
        """
        # Grab the raw bytes and push it into the thread safe queue.
        data = audio.get_raw_data()
        data_queue.put(data)

    # Create a background thread that will pass us raw audio bytes.
    # We could do this manually but SpeechRecognizer provides a nice helper.
    recorder.listen_in_background(source, record_callback, phrase_time_limit=record_timeout)

    # Cue the user that we're ready to go.
    print("Model loaded.\n")
    transcription_in_progress = True

    while transcription_in_progress:
        try:
            now = datetime.utcnow()
            # Pull raw recorded audio from the queue.
            if not data_queue.empty():
                phrase_complete = False
                # If enough time has passed between recordings, consider the phrase complete.
                # Clear the current working audio buffer to start over with the new data.
                if phrase_time and now - phrase_time > timedelta(seconds=phrase_timeout):
                    last_sample = bytes()
                    phrase_complete = True
                # This is the last time we received new audio data from the queue.
                phrase_time = now

                # Concatenate our current audio data with the latest audio data.
                while not data_queue.empty():
                    data = data_queue.get()
                    last_sample += data

                # Use AudioData to convert the raw data to wav data.
                audio_data = sr.AudioData(last_sample, source.SAMPLE_RATE, source.SAMPLE_WIDTH)
                wav_data = io.BytesIO(audio_data.get_wav_data())

                time_stamp = int(time.time())
                file_name = f"./logs/{time_stamp}_audio_recording.wav"
                # Write wav data to the temporary file as bytes.
                with open(file_name, 'w+b') as f:
                    f.write(wav_data.read())

                # Read the transcription.
                result = audio_model.transcribe(file_name, fp16=torch.cuda.is_available())
                text = result['text'].strip()

                # If we detected a pause between recordings, add a new item to our transcripion.
                # Otherwise edit the existing one.
                if phrase_complete:
                    print(f"Transcribed text: {text}")
                    transcription.append(text)
                else:
                    transcription[-1] = text
                
                sleep(0.25)
        except KeyboardInterrupt:
            break


import torch
import os
import sys

from pathlib import Path
p = Path(__file__).parents[3]

sys.path.append(str(p))
from utils import *
from model import Adapter, Adapter_FC

def load_pretrained_mb_and_adapters(config=None, memory_bank_v_path=None, memory_bank_t_path=None, adapter_type=None, adapter_weights_path=None):
    """Returns the pretrained visual embeddings and textual embeddings from the memory bank and the pretrained query adapter"""
    if config:
        with torch.no_grad():
            model_dir_root = get_model_dir_root(config)
            model_dir = f"{model_dir_root}/alpha-beta/{config['alpha']}-{config['beta']}"
            model_prefix = f"best_lr_{config['lr']}_aug_{config['augment_epoch']}_epochs_{config['train_epoch']}"

            best_model_path_v = os.path.join(model_dir, f"{model_prefix}_v.pt")
            best_model_path_t = os.path.join(model_dir, f"{model_prefix}_t.pt")
            best_model_path_a = os.path.join(model_dir, f"{model_prefix}_a.pt")

            try:
                embeddings_v = torch.load(best_model_path_v)
                embeddings_t = torch.load(best_model_path_t)
            except:
                raise FileNotFoundError(f"File does not exist: {best_model_path_v} and {best_model_path_t}")

            NxK, ndim = embeddings_v.shape

            if 'conv' in config['adapter']:
                adapter = Adapter(ndim, c_type=config['adapter'], dtype=torch.half).cuda()
            elif config['adapter'] == 'fc':
                adapter = Adapter_FC(ndim, dtype=torch.half).cuda()

            try:
                adapter.load_state_dict(torch.load(best_model_path_a))
            except:
                raise FileNotFoundError(f"File does not exist: {best_model_path_a}")
    else:
        with torch.no_grad():
            try:
                embeddings_v = torch.load(memory_bank_v_path)
                embeddings_t = torch.load(memory_bank_t_path)
            except:
                raise FileNotFoundError(f"File does not exist: {memory_bank_v_path} and {memory_bank_t_path}")

            NxK, ndim = embeddings_v.shape

            if adapter_type==None:
                raise Exception("Please mention the adapter type in the args or in the config file.")
                
            if 'conv' in adapter_type:
                adapter = Adapter(ndim, c_type=config['adapter'], dtype=torch.half).cuda()
            elif adapter_type == 'fc':
                adapter = Adapter_FC(ndim, dtype=torch.half).cuda()

            try:

                adapter_state_dict = torch.load(adapter_weights_path)
            except:
                raise FileNotFoundError(f"File does not exist: {best_model_path_a}")
            
            adapter.load_state_dict(adapter_state_dict)
    
    return embeddings_v, embeddings_t, adapter

def pre_load_features_without_cache(clip_model, loader):

    """Returns the clip features for images without caching them."""
    features = []
    with torch.no_grad():
        for i, images in enumerate(tqdm(loader)):
            images= images.cuda()
            image_features = clip_model.encode_image(images)
            image_features /= image_features.norm(dim=-1, keepdim=True)
            features.append(image_features)

    features = torch.cat(features)

    return features




from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import random
import clip
import numpy as np
import torch
import sys

from .model_utils import load_pretrained_mb_and_adapters
from torchvision.datasets import ImageFolder

from pathlib import Path
p = Path(__file__).parents[3]

sys.path.append(str(p))
from utils import *
from imagenetv2_pytorch import ImageNetV2Dataset
from PIL import Image
from torch.utils.data import Dataset



class ImageNetSketchDataset(Dataset):
    def __init__(self, root_dir, transform=None):
        self.root_dir = root_dir
        self.transform = transform
        self.classes = sorted(os.listdir(root_dir))
        self.class_to_idx = {cls_name: idx for idx, cls_name in enumerate(self.classes)}
        self.images = self._load_images()

    def __len__(self):
        return len(self.images)

    def __getitem__(self, index):
        image_path, label = self.images[index]
        image = Image.open(image_path).convert("RGB")

        if self.transform is not None:
            image = self.transform(image)

        return image, label

    def _load_images(self):
        images = []
        for cls_name in self.classes:
            cls_dir = os.path.join(self.root_dir, cls_name)
            if os.path.isdir(cls_dir):
                label = self.class_to_idx[cls_name]
                for filename in os.listdir(cls_dir):
                    image_path = os.path.join(cls_dir, filename)
                    images.append((image_path, label))
        return images


def test_ood_performance(cfg, test_dataset_name, n_workers, test_bs, memory_bank_v_path=None, memory_bank_t_path=None, adapter_type=None, adapter_weights_path=None):

    # CLIP
    clip_model, preprocess = clip.load(cfg['backbone'])
    clip_model.eval()

    # Prepare dataset; SEED is fetched from utils.py
    seed = get_seed()
    random.seed(seed)
    np.random.seed(seed)
    g = torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    print("Preparing dataset.")
    if test_dataset_name=="imagenet_v2":
        test_dataset = ImageNetV2Dataset("matched-frequency", transform=preprocess)
        test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=test_bs, num_workers=n_workers, shuffle=False)
    elif test_dataset_name=="imagenet_sketch":
        test_dataset = ImageFolder("./DATA/sketch", transform=preprocess)
        test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=test_bs, num_workers=n_workers, shuffle=False)


    # Pre-load test features
    # print("\nLoading visual features and labels from test set.")
    test_features, test_labels = pre_load_features(cfg, "test", clip_model, test_loader)

    with torch.no_grad():
        print("Testing...")

        embeddings_v, embeddings_t, adapter = load_pretrained_mb_and_adapters(memory_bank_v_path=memory_bank_v_path, 
                                                                            memory_bank_t_path=memory_bank_t_path, 
                                                                            adapter_type=adapter_type, 
                                                                            adapter_weights_path=adapter_weights_path)
        NxK, ndim = embeddings_v.shape
        K = cfg['shots']
        N = NxK//K
        print(K, N)

        zs_imgs = embeddings_v.view(-1, K, ndim)
        zs_imgs = zs_imgs / zs_imgs.norm(dim=-1, keepdim=True)
        z_img_proto = zs_imgs.mean(dim=1)
        z_img_proto = z_img_proto / z_img_proto.norm(dim=-1, keepdim=True)

        zs_text = embeddings_t
        z_text_proto = zs_text / zs_text.norm(dim=-1, keepdim=True)

        test_features = adapter(test_features)
        test_features = test_features / test_features.norm(dim=-1, keepdim=True)

        p = P(test_features, z_img_proto, z_text_proto, cfg['alpha'], cfg['beta'])

        test_acc = (p.max(1)[1] == test_labels).float().mean() * 100.0

        return test_acc


from flair.data import Sentence
from flair.models import SequenceTagger

class VerbAndNounTagger:
    def __init__(self, verb_dictionary_path, noun_dictionary_path):

        """Load the flair sequence tagger and the allowed verb and noun dictionary."""
        self.tagger = SequenceTagger.load("flair/pos-english")

        verb_file = open(verb_dictionary_path, "r")
        verb_list = verb_file.readlines()
        self.allowed_verb_set = set([x.strip("\n") for x in verb_list])


        noun_file = open(noun_dictionary_path, "r")
        noun_list = noun_file.readlines()
        self.allowed_noun_set = set([x.replace("_", " ").strip("\n") for x in noun_list])
    
    def tag_sentence(self, text):
        """Tag a sentence with its POS and return the noun and verb from the sentence if they are present in the dictionary."""
        sentence = Sentence(text)
        self.tagger.predict(sentence)
        
        word_tag_list = []

        for entity in sentence.get_labels():
            # Uncomment this if you want to debug the POS tag for each word you speak.
            # print(entity)
            word = entity.shortstring.split("/")[0].strip("\"").lower()
            word_tag_list.append((word, entity.value))
        
        return word_tag_list

    def find_valid_noun_and_verb(self, text):
        """Tag a sentence with its POS and return the noun and verb from the sentence if they are present in the dictionary."""
        word_tag_list = self.tag_sentence(text)
        parsed_verb = None
        parsed_noun = None
        
        idx = 0
        while idx < len(word_tag_list):
            curr_word, curr_tag = word_tag_list[idx]

            # For nouns(objects) like mustard bottle, we need to detect both the NN's together. 
            # Therefore, we move ahead of the array and combine subsequences with same tags into a single word.
            while idx + 1 < len(word_tag_list) and curr_tag==word_tag_list[idx+1][1]:
                curr_word += " " + word_tag_list[idx+1][0]
                idx += 1

            if (curr_tag=="NN" or curr_tag=="NNP" or curr_tag=="NNS") and curr_word in self.allowed_noun_set:
                parsed_noun = curr_word
            if curr_tag=="VB" and curr_word in self.allowed_verb_set:
                parsed_verb = curr_word

            idx += 1
        
        return parsed_verb, parsed_noun


from pathlib import Path
import setuptools
from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
required = (this_directory / "requirements.txt").read_text().split("\n")


setup(name='proto_clip_toolkit', 
      description="A simple toolkit from Proto-CLIP demo that provies speech recognition, part-of-speech tagging and realworld robot demo APIs.",
      author="IRVL-UTD: Intelligent Robotics and Vision Lab at the University of Texas at Dallas",
      author_email="irvl.utd@gmail.com",
      version='0.2',
      install_requires=required,
      url="https://github.com/IRVLUTD/Proto-CLIP",
      license="MIT",
      packages=setuptools.find_packages(),
      include_package_data=True,
      python_requires=">=3.7",
      long_description=long_description,
    long_description_content_type='text/markdown'
)



import torch
import os
import matplotlib.pyplot as plt
import argparse
import yaml
import sys

from pathlib import Path
p = Path(__file__).parents[3]

sys.path.append(str(p))
from proto_datasets import build_dataset
from sklearn.manifold import TSNE
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
from utils import build_cache_model, get_textual_memory_bank
import json
import cv2

def get_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', dest='config', help='settings of Proto-CLIP in yaml format', required=True)
    parser.add_argument('--splits_path', dest='splits_path', help='path to the splits file for the particular dataset', required=True)
    parser.add_argument('--memory_bank_v_path', dest='memory_bank_v_path', help='path to the visual embeddings memory bank', required=True)
    parser.add_argument('--memory_bank_t_path', dest='memory_bank_t_path', help='path to the textual embeddings memory bank', required=True)
    parser.add_argument('--after_train', dest='after_train', help='save embeddings after training', action='store_true')
    args = parser.parse_args()
    return args


def parse_splits_file(config_path):
    """Returns the map class id to the class name using the splits json file."""

    class_id_mapping = {}
    f = open(config_path)
    data = json.load(f)

    for config_data in data["train"]:
        class_id_mapping[config_data[1]] = config_data[2]
    
    return class_id_mapping

def get_image_samples(txt_path, n_classes):
    """Returns a single representative image sample for each class."""
    f = open(txt_path, "r")
    data = f.readlines()

    data = [x.strip("\n") for x in data]
    output_image_locations = []

    for i in range(n_classes):
        #We pick the first sample of the support set for a given class.
        img_idx = i*16

        output_image_locations.append(data[img_idx])

    return output_image_locations

def get_tsne_coordinates(z_img_proto, z_text_proto, n_class):
    """Returns the 2-dimensional t-SNE coordinates for the image and the text embeddings."""
    X = torch.vstack((
        z_img_proto,
        z_text_proto,
        # zq_imgs.view(-1, zq_imgs.shape[-1])
    )).cpu().data.numpy()
    tsne_X = TSNE(n_components=2, perplexity=10, random_state=1).fit_transform(X)
    zi, zt = tsne_X[:n_class], tsne_X[n_class: ]

    return zi, zt

def plot_tsne_after(z_img_proto, z_text_proto, txt_prompts):
    """Returns the t-SNE plot for the visual and textual embeddings after training."""
    n_class = z_img_proto.shape[0]
    zi, zt = get_tsne_coordinates(z_img_proto, z_text_proto, n_class)

    image_locations = get_image_samples("./image_locations.txt", n_class)
    _, ax = plt.subplots(figsize=(50, 50))
    fontsize = 10
    for idx, (x, y) in enumerate(zip(zi[:, 0], zi[:, 1])):
        img = plt.imread(image_locations[idx])
        img = cv2.resize(img, (48, 48))

        imagebox = OffsetImage(img)  # Adjust the zoom level as desired
        ab = AnnotationBbox(imagebox, (x, y), frameon=False, zorder=1)
        
        # ax.set_aspect('equal')
        ax.scatter(x, y, zorder=4, s=32, c='cyan', marker=".")
        ax.add_artist(ab)
        ax.annotate(
            txt_prompts[idx], xy=(x, y + 1), ha='center', c="crimson", fontsize=fontsize)
    
    ax.scatter(zt[:, 0], zt[:, 1], c='aquamarine', zorder=3, marker="+", s=128)

    ax.axis('off')

    plt.savefig("./test_plot_after.png", dpi=300)

def plot_tsne_before(z_img_proto, z_text_proto, txt_prompts):
    """Returns the t-SNE plot for the visual and textual embeddings before training."""
    n_class = z_img_proto.shape[0]
    zi, zt = get_tsne_coordinates(z_img_proto, z_text_proto, n_class)

    image_locations = get_image_samples("./image_locations.txt", n_class)
    _, ax = plt.subplots(figsize=(50, 50))
    fontsize = 25
    for idx, (x, y) in enumerate(zip(zi[:, 0], zi[:, 1])):
        img = plt.imread(image_locations[idx])
        img = cv2.resize(img, (48, 48))

        imagebox = OffsetImage(img)  # Adjust the zoom level as desired
        ab = AnnotationBbox(imagebox, (x, y), frameon=False, zorder=1)
        
        ax.scatter(x, y, zorder=4, s=32, c='cyan', marker=".")
        ax.add_artist(ab)
    
    ax.scatter(zt[:, 0], zt[:, 1], c='lightseagreen', zorder=3, marker="P", s=128)


    for i in range(len(txt_prompts)):    
        ax.annotate(
            txt_prompts[i], (zt[i, 0], zt[i, 1] + 0.2), c='crimson', fontsize=fontsize)

    ax.axis('off')

    plt.savefig("./test_plot_before.png", dpi=300)

if __name__=="__main__":
    args = get_arguments()

    cfg = yaml.load(open(args.config, 'r'), Loader=yaml.Loader)

    dataset = build_dataset(cfg['dataset'], cfg['root_path'], cfg['shots'])

    if not args.after_train:
        #This function will not work if you do not have the cache available.
        #You would need to create the cache and place the initial embeddings inside it. 
        #An easy way to do it is to run the main.py for 1 epoch.
        visual_memory_keys, visual_mem_values = build_cache_model(cfg, None, None)

        text_prompts, textual_memory_bank = get_textual_memory_bank(cfg, [], None, None)
        
        embeddings_v = visual_memory_keys.t()
        embeddings_t = textual_memory_bank.t()
    else:
        best_model_path_v = args.memory_bank_v_path
        best_model_path_t = args.memory_bank_t_path

        try:
            embeddings_v = torch.load(best_model_path_v)
            embeddings_t = torch.load(best_model_path_t)
        except:
            raise FileNotFoundError(f"File does not exist: {best_model_path_v} and {best_model_path_t}") 
    
    
    NxK, ndim= embeddings_v.shape
    K = cfg['shots']
    N = NxK//K

    zs_imgs = embeddings_v.view(-1, K, ndim)
    zs_imgs = zs_imgs / zs_imgs.norm(dim=-1, keepdim=True)
    z_img_proto = zs_imgs.mean(dim=1)
    z_img_proto = z_img_proto / z_img_proto.norm(dim=-1, keepdim=True)
    
    zs_text = embeddings_t
    z_text_proto = zs_text / zs_text.norm(dim=-1, keepdim=True)
        
    class_id_mapping = parse_splits_file(args.splits_path)

    if args.after_train:
        plot_tsne_after(z_img_proto, z_text_proto, list(class_id_mapping.values()))
    else:
        plot_tsne_before(z_img_proto, z_text_proto, list(class_id_mapping.values()))